@@ -1,12 +1,12 @@
-(* $Id: ag_biniou.ml 52210 2010-11-22 09:50:55Z martin $ *)
-
 (*
   Mapping from ATD to biniou
 *)
 
-type biniou_int = 
+type biniou_int =
     [ `Svint | `Uvint | `Int8 | `Int16 | `Int32 | `Int64 ]
 
+type biniou_float = [ `Float32 | `Float64 ]
+
 type biniou_list = [ `Array | `Table ]
 
 type biniou_field = { biniou_unwrapped : bool }
@@ -16,7 +16,7 @@ type biniou_repr =
     | `Unit
     | `Bool
     | `Int of biniou_int
-    | `Float
+    | `Float of biniou_float
 
     | `String
     | `Sum
@@ -24,7 +24,9 @@ type biniou_repr =
     | `Tuple
     | `List of biniou_list
     | `Option
+    | `Nullable
     | `Shared
+    | `Wrap
     | `External
 
     | `Cell
@@ -43,6 +45,12 @@ let biniou_int_of_string s : biniou_int option =
     | "int64" -> Some `Int64
     | _ -> None
 
+let biniou_float_of_string s : biniou_float option =
+  match s with
+      "float32" -> Some `Float32
+    | "float64" -> Some `Float64
+    | _ -> None
+
 let biniou_list_of_string s : biniou_list option =
   match s with
       "array" -> Some `Array
@@ -52,5 +60,8 @@ let biniou_list_of_string s : biniou_list option =
 let get_biniou_int an =
   Atd_annot.get_field biniou_int_of_string `Svint ["biniou"] "repr" an
 
+let get_biniou_float an =
+  Atd_annot.get_field biniou_float_of_string `Float64 ["biniou"] "repr" an
+
 let get_biniou_list an =
   Atd_annot.get_field biniou_list_of_string `Array ["biniou"] "repr" an
