@@ -1,4 +1,3 @@
-(* $Id: ag_ob_run.ml 51992 2010-11-18 07:40:37Z martin $ *)
 
 (*
   Runtime library
@@ -27,7 +26,7 @@ let unsupported_variant h has_arg =
 
 
 let missing_tuple_fields len req_fields =
-  let missing = 
+  let missing =
     List.fold_right (
       fun i acc -> if i >= len then i :: acc else acc
     ) req_fields []
@@ -142,18 +141,34 @@ let read_int64 ib =
   else
     read_error_at ib
 
-let get_float_reader tag =
+let get_float32_reader tag =
+  if tag = Bi_io.float32_tag then
+    Bi_io.read_untagged_float32
+  else
+    tag_error tag "float32"
+
+let get_float64_reader tag =
   if tag = Bi_io.float64_tag then
     Bi_io.read_untagged_float64
   else
-    tag_error tag "float"
+    tag_error tag "float64"
 
-let read_float ib =
+let get_float_reader = get_float64_reader
+
+let read_float32 ib =
+  if Bi_io.read_tag ib = Bi_io.float32_tag then
+    Bi_io.read_untagged_float32 ib
+  else
+    read_error_at ib
+
+let read_float64 ib =
   if Bi_io.read_tag ib = Bi_io.float64_tag then
     Bi_io.read_untagged_float64 ib
   else
     read_error_at ib
 
+let read_float = read_float64
+
 let get_string_reader tag =
   if tag = Bi_io.string_tag then
     Bi_io.read_untagged_string
@@ -213,7 +228,7 @@ let write_tagged tag write buf x =
 let write_untagged_option write buf x =
   match x with
       None -> Bi_io.write_numtag buf 0 false
-    | Some x -> 
+    | Some x ->
 	Bi_io.write_numtag buf 0 true;
 	write buf x
 
