@@ -51,38 +51,42 @@ let write_array write_item ob a =
 let write_assoc_list write_item ob l =
   Bi_outbuf.add_char ob '{';
   list_iter (
-    fun ob (k, v) -> 
+    fun ob (k, v) ->
       Yojson.Safe.write_string ob k;
       Bi_outbuf.add_char ob ':';
       write_item ob v
   ) write_comma ob l;
   Bi_outbuf.add_char ob '}'
-      
+
 let write_assoc_array write_item ob l =
   Bi_outbuf.add_char ob '{';
   array_iter (
-    fun ob (k, v) -> 
+    fun ob (k, v) ->
       Yojson.Safe.write_string ob k;
       Bi_outbuf.add_char ob ':';
       write_item ob v
   ) write_comma ob l;
   Bi_outbuf.add_char ob '}'
-      
+
 
 let write_option write_item ob = function
     None -> Bi_outbuf.add_string ob "<\"None\">"
-  | Some x -> 
+  | Some x ->
       Bi_outbuf.add_string ob "<\"Some\":";
       write_item ob x;
       Bi_outbuf.add_string ob ">"
 
 let write_std_option write_item ob = function
     None -> Bi_outbuf.add_string ob "\"None\""
-  | Some x -> 
+  | Some x ->
       Bi_outbuf.add_string ob "[\"Some\",";
       write_item ob x;
       Bi_outbuf.add_string ob "]"
 
+let write_nullable write_item ob = function
+    None -> Bi_outbuf.add_string ob "null"
+  | Some x -> write_item ob x
+
 let write_int8 ob x =
   Yojson.Safe.write_int ob (int_of_char x)
 
@@ -92,6 +96,21 @@ let write_int32 ob x =
 let write_int64 ob x =
   Bi_outbuf.add_string ob (Int64.to_string x)
 
+let min_float = float min_int
+let max_float = float max_int
+
+let write_float_as_int ob x =
+  if x >= min_float && x <= max_float then
+    Yojson.Safe.write_int ob
+      (int_of_float (if x < 0. then x -. 0.5 else x +. 0.5))
+  else
+    match classify_float x with
+        FP_normal
+      | FP_subnormal
+      | FP_zero -> Bi_outbuf.add_string ob (Printf.sprintf "%.0f" x)
+      | FP_infinite -> error "Cannot convert inf or -inf into a JSON int"
+      | FP_nan -> error "Cannot convert NaN into a JSON int"
+
 let read_null p lb =
   Yojson.Safe.read_space p lb;
   Yojson.Safe.read_null p lb
@@ -162,7 +181,7 @@ let read_until_field_value p lb =
   Yojson.Safe.read_space p lb
 
 let missing_tuple_fields len req_fields =
-  let missing = 
+  let missing =
     List.fold_right (
       fun i acc -> if i >= len then i :: acc else acc
     ) req_fields []
