@@ -1,5 +1,3 @@
-(* $Id: test_atdgen_main.ml 53195 2011-01-28 23:38:12Z martin $ *)
-
 open Printf
 
 let current_section = ref ""
@@ -23,13 +21,25 @@ let fail () =
 let check b =
   if not b then fail ()
 
+let check_valid = function
+  | None -> ()
+  | Some error ->
+      printf "%s\n%!" (Ag_util.Validation.string_of_error error);
+      fail ()
+
+let check_invalid = function
+  | None -> fail ()
+  | Some error ->
+      printf "%s\n%!" (Ag_util.Validation.string_of_error error)
+
+
 let expect_error f x =
-  try 
+  try
     ignore (f x);
     printf "Did not get expected error\n%!";
     fail ()
   with
-      Ag_ob_run.Error s 
+      Ag_ob_run.Error s
     | Ag_oj_run.Error s ->
         printf "Got expected error:\n%s\n%!" s
 
@@ -69,7 +79,7 @@ let test_biniou_missing_field () =
 
 let test_biniou_missing_cell () =
   section "biniou missing tuple fields";
-  expect_error 
+  expect_error
     Test.extended_tuple_of_string
     (Test.string_of_base_tuple test_missing_tuple)
 
@@ -80,7 +90,7 @@ let test_json_missing_field () =
 
 let test_json_missing_cell () =
   section "json missing tuple fields";
-  expect_error 
+  expect_error
     Testj.extended_tuple_of_string
     (Testj.string_of_base_tuple test_missing_tuple)
 
@@ -110,10 +120,34 @@ let test_json_assoc_array () =
   f [| ("a", 0) |];
   f [| ("a", 0); ("b", 1) |]
 
+let test_json_int_ocaml_float_gen of_json to_json kind () =
+  section ("json ints derived from ocaml floats: " ^ kind);
+  let l1 = [0.; 0.1; -0.1; 0.6; -0.6] in
+  check (of_json (to_json l1) = [0.; 0.; 0.; 1.; -1.]);
+
+  let l2 = [ 12345678901234567890.; -12345678901234567890. ] in
+  let l2' = of_json (to_json l2) in
+  List.map2 (fun x x' -> abs_float (1. -. x /. x') < 1e-15) l2 l2';
+
+  expect_error to_json [infinity];
+  expect_error to_json [neg_infinity];
+  expect_error to_json [nan]
+
+let test_json_int_ocaml_float () =
+  test_json_int_ocaml_float_gen
+    Test3j.unixtime_list_of_string
+    Test3j.string_of_unixtime_list
+    "int <ocaml repr=\"float\">" ();
+  test_json_int_ocaml_float_gen
+    Testj.unixtime_list_of_string
+    Testj.string_of_unixtime_list
+    "float <json repr=\"int\">" ()
+
+
 
 let make_mixed_record_array n =
   Array.init n (
-    fun i -> 
+    fun i ->
       {
 	Test.field0 = Some i;
 	field1 = Some 0.555;
@@ -136,7 +170,8 @@ let make_mixed_record_array n =
 	field10 = true;
 	field11 = false;
         field12 = [ (); () ];
-        field13 = [ Some "abc"; None ]
+        field13 = [ Some "abc"; None ];
+        field14 = (2012, Some 3, None);
       }
   )
 
@@ -175,7 +210,8 @@ let test_correctness_data = {
       field10 = true;
       field11 = false;
       field12 = [ (); () ];
-      field13 = [ Some "abc"; None ]
+      field13 = [ Some "abc"; None ];
+      field14 = (2012, Some 3, None);
     }
   ];
   x4 = 0x0807060504030201L;
@@ -245,12 +281,16 @@ let test_json_space () =
   ignore (Testj.test_of_string pp)
 
 
+let test_validators0 () =
+  section "validators0";
+  check_valid (Testv.validate_test [] test_correctness_data)
+
 let test_validators1 () =
   section "validators1";
   let valid = (0, 1.) in
   let invalid = (1, 0.) in
-  check (Testv.validate_base_tuple valid);
-  check (not (Testv.validate_base_tuple invalid));
+  check_valid (Testv.validate_base_tuple [] valid);
+  check_invalid (Testv.validate_base_tuple [] invalid);
 
   let x1 = {
     Test.b0x = 1;
@@ -261,29 +301,29 @@ let test_validators1 () =
     b5x = 1.1;
   }
   in
-  check (not (Testv.validate_extended x1));
+  check_invalid (Testv.validate_extended [] x1);
   let x2 = { x1 with Test.b1x = false } in
-  check (Testv.validate_extended x2);
+  check_valid (Testv.validate_extended [] x2);
   let x3 = { x2 with Test.b0x = -1 } in
-  check (not (Testv.validate_extended x3))
+  check_invalid (Testv.validate_extended [] x3)
 
 let test_validators2 () =
   section "validators2";
   let v1 = `A in
-  check (not (Testv.validate_p v1));
+  check_invalid (Testv.validate_p [] v1);
   let v2 = `B { Test.a = 0; b = true; c = `C } in
-  check (Testv.validate_p v2)
+  check_valid (Testv.validate_p [] v2)
 
 let test_validators3 () =
   section "validators3";
   let o = Some 0 in
-  check (not (Testv.validate_option_validation o))
+  check_invalid (Testv.validate_option_validation [] o)
 
 let test_validators4 () =
   section "validators4";
   let x = { Test.val2_x = { Test.val1_x = 0 };
             val2_y = Some { Test.val1_x = 1 } } in
-  check (not (Testv.validate_val2 x))
+  check_invalid (Testv.validate_val2 [] x)
 
 let test_json_files () =
   section "json files";
@@ -318,7 +358,7 @@ let test_raw_json () =
   let s = Test3j.string_of_t x in
   let x' = Test3j.t_of_string s in
   check (x = x')
-  
+
 let test_biniou_sharing_graph () =
   section "biniou sharing - graph";
   let a = { Test3b.value = "a"; neighbors = [] } in
@@ -341,7 +381,7 @@ let test_biniou_sharing_graph () =
   in
   check (a' == a'');
   check (b' == b'')
-  
+
 let test_biniou_sharing_strings () =
   section "biniou sharing - strings";
   let x = ref "abc" in
@@ -355,6 +395,58 @@ let test_biniou_sharing_strings () =
   check (a'.(0) == a'.(1));
   check (a'.(0) != b')
 
+let test_wrapping_ints () =
+  section "ocaml wrapping - ints";
+  let x = Test_lib.Natural.wrap 7 in
+  let json = Testj.string_of_natural x in
+  let x' = Testj.natural_of_string json in
+  check (x = x');
+
+  let biniou = Test.string_of_natural x in
+  let x'' = Test.natural_of_string biniou in
+  check (x = x'');
+
+  try ignore (Testj.natural_of_string "-1"); check false
+  with Failure _ -> ()
+
+let test_double_wrapping () =
+  section "ocaml wrapping - double wrapping";
+  let x = Test_lib.Even_natural.wrap (Test_lib.Natural.wrap 10) in
+  let json = Testj.string_of_even_natural x in
+  let x' = Testj.even_natural_of_string json in
+  check (x = x')
+
+let test_wrapping_with_validation () =
+  section "ocaml wrapping - with validation";
+  let x = `Id "" in
+  try ignore (Testv.validate_id [] x); check false
+  with Failure "empty" -> ()
+
+let test_ignored_wrap () =
+  section "ocaml wrapping - wrap constructor without wrapper";
+  let x = { Test.some_field = 0 } in
+  try ignore (Testv.validate_no_real_wrap [] x); check false
+  with Failure "passed" -> ()
+
+let test_biniou_float32 () =
+  section "check length of floats serialized as float32";
+  let x = { Test.f32 = 1.23456789; Test.f64 = 1.98765432 } in
+  let s = Test.string_of_floats x in
+  let x' = Test.floats_of_string s in
+  check Test.(abs_float (x.f32 -. x'.f32) < 1e-6);
+  check (String.length s = 24)
+
+let test_json_float_decimals () =
+  section "print JSON floats with maximum number of decimal places";
+  let x = {
+    Testj.sqrt2_5 = sqrt 2.;
+    small_2 = 0.000123456789;
+    large_2 = 1234567890123.;
+  } in
+  let s = Testj.string_of_precision x in
+  print_endline s;
+  check (s = "{\"sqrt2_5\":1.4142,\"small_2\":0.00012,\"large_2\":1.2e+12}")
+
 let all_tests = [
   test_ocaml_internals;
   test_biniou_missing_field;
@@ -364,9 +456,11 @@ let all_tests = [
   test_json_extra_field_warning;
   test_json_assoc_list;
   test_json_assoc_array;
+  test_json_int_ocaml_float;
   test_biniou_correctness;
   test_json_correctness;
   test_json_space;
+  test_validators0;
   test_validators1;
   test_validators2;
   test_validators3;
@@ -376,6 +470,12 @@ let all_tests = [
   test_raw_json;
   test_biniou_sharing_graph;
   test_biniou_sharing_strings;
+  test_wrapping_ints;
+  test_double_wrapping;
+  test_wrapping_with_validation;
+  test_ignored_wrap;
+  test_biniou_float32;
+  test_json_float_decimals;
 ]
 
 let quality_test () =
